# auxiliary R functions for motif discovery

library(Biostrings)
library(dplyr) # distinct rows

#' Create train and test sequence files
#'
#' @param term A GO term sequence file name 
#' @param trainSize The proportion of train samples
#'
#' @return void
#' @export
splitSequences <- function(term,trainSize){
  seqDatabase <- readAAStringSet(paste0("CAFA3_training_data/",term,".fasta"))
  trSet<-sample(length(seqDatabase))[1:trunc(trainSize*length(seqDatabase))]
  writeXStringSet(seqDatabase[trSet,],paste0("CAFA3_training_data/",term,"_train.fasta"))
  writeXStringSet(seqDatabase[!(seq(length(seqDatabase)) %in% trSet),],paste0("CAFA3_training_data/",term,"_test.fasta"))
}

#' Discover motifs in train files in all the groups
#' Score the found motifs in all test files
#' Results stored in external files
#' 
#' @param GOlist A vector of GO term sequence file names 
#'
#' @return void
#' @export
findAndScore <- function(GOlist){
  for (term in GOlist){
    # negative term selected randomly from the rest of the GO term list
    # used as a reference when searching for motifs
    negative <- sample(GOlist[GOlist!=term])[1]
    system2(command = "streme",
            args = c(paste0("-p CAFA3_training_data/",term,"_train.fasta"),
                     paste0("-n CAFA3_training_data/",negative,"_train.fasta"),
                     paste0("-o streme_out/",term,"_motif"),
                     "--protein","-thres 0.1","-time 120")
    )
    for (eterm in GOlist){
      system2(command = "fimo",
              args = c(paste0("--oc fimo_out/",term, "_", eterm,"_motif"),
                       paste0("streme_out/",term,"_motif/streme.txt"),
                       paste0("CAFA3_training_data/",eterm,"_test.fasta")
              )
      )
    }  
  }
}  

#' Get a score file from FIMO res file
#'
#' @param resName A FIMO res file name (string expected)
#' @param negThres A negative score to be used when no motif match found
#'
#' @return df A score data frame (sequences in rows, motifs in columns)
#' @export
procFIMO <- function(resName,negThres){
  res<-read.csv(resName,sep="\t")
  res<-head(res, -3) # remove the last three descriptive lines in res file
  motifs<-unique(res$motif_id)
  df<-data.frame(matrix(ncol = length(motifs), nrow = 0))
  colnames(df) <- motifs
  seqs<-unique(res$sequence_name)
  for (seq in seqs){
    r<-res[res$sequence_name==seq,]
    r<-aggregate(x = r$score, by = list(r$motif_id), FUN = max)
    df.r<-rep(negThres,length(motifs))
    df.r[match(r[,1],names(df))]<-r[,2]
    df[nrow(df)+1,]<-df.r
  }
  df$Seqs<-seqs
  return(df)
}

#' Get a joint score file from FIMO res files for one term
#'
#' @param term A GO term
#' @param negThres A negative score to be used when no motif match found
#'
#' @return df A score data frame (sequences in rows, motifs in columns)
#' @export
procFIMOterm <- function(term,negThres){
  dfi<-procFIMO(paste0("fimo_out/",term, "_", term,"_motif/fimo.tsv"),negThres)
  for (eterm in GOlist[GOlist!=term]){
    dfa<-procFIMO(paste0("fimo_out/",eterm, "_", term,"_motif/fimo.tsv"),negThres)
    dfi<-merge(dfi,dfa,by="Seqs",all=T)
  }
  dfi$class<-term
  ## remove duplicate proteins (having more GO annotations we work with, e.g., more child terms merged)
  return(dfi %>% distinct(Seqs,.keep_all= TRUE))
}
